SystemOrganization addCategory: #TelnetSessionBasic!

SharedPool subclass: #TelnetCmdMatchLib
	instanceVariableNames: ''
	classVariableNames: 'CiscoEnaPromptRxp CiscoLoginRxp CiscoOpPromptRxp CiscoPasswordRxp JunCustomPromptRxp JunCustomPromptStr JunLoginRxp JunOpPromptRxp JunPasswordRxp RpcReplyStr XmlJunoscriptMetaSessionRxp XmlMetaRxp'
	poolDictionaries: ''
	category: 'TelnetSessionBasic'!

!TelnetCmdMatchLib commentStamp: '<historical>' prior: 0!
Strings to match against CLI prompts of network equipment!

!TelnetCmdMatchLib class methodsFor: 'initialization' stamp: 'DanilOsipchuk 6/18/2013 17:00'!
initialize
	"self initialize"
	
	JunLoginRxp := 'login\:.*' asRegexIgnoringCase.
	JunPasswordRxp := 'password\:.*' asRegexIgnoringCase.
	JunOpPromptRxp := '.*@.*>.*' asRegex.
	JunCustomPromptStr := '````TelnetSessionBasic~~>'.
	JunCustomPromptRxp := JunCustomPromptStr asRegex.
	CiscoLoginRxp := 'username\:.*' asRegexIgnoringCase.
	CiscoPasswordRxp := 'password\:.*' asRegexIgnoringCase.
	CiscoOpPromptRxp := '.*>' asRegex.
	CiscoEnaPromptRxp := '.*#' asRegex.
	RpcReplyStr := '</rpc-reply>'.
	XmlMetaRxp := '<!!--.*-->' asRegex.
	XmlJunoscriptMetaSessionRxp := '<!!-- user.* -->' asRegex.
	! !

SharedPool subclass: #TelnetConstantsRFC854
	instanceVariableNames: ''
	classVariableNames: 'AO AYT BRK DM DO DONT EC EL GA IAC IP NOP SB SE SGA WILL WONT'
	poolDictionaries: ''
	category: 'TelnetSessionBasic'!

!TelnetConstantsRFC854 commentStamp: '<historical>' prior: 0!
Straight from RFC854

NAME               CODE              MEANING

      SE                  240    End of subnegotiation parameters.
      NOP                 241    No operation.
      Data Mark           242    The data stream portion of a Synch.
                                 This should always be accompanied
                                 by a TCP Urgent notification.
      Break               243    NVT character BRK.
      Interrupt Process   244    The function IP.
      Abort output        245    The function AO.
      Are You There       246    The function AYT.
      Erase character     247    The function EC.
      Erase Line          248    The function EL.
      Go ahead            249    The GA signal.
      SB                  250    Indicates that what follows is
                                 subnegotiation of the indicated
                                 option.
      WILL (option code)  251    Indicates the desire to begin
                                 performing, or confirmation that
                                 you are now performing, the
                                 indicated option.
      WON'T (option code) 252    Indicates the refusal to perform,
                                 or continue performing, the
                                 indicated option.
      DO (option code)    253    Indicates the request that the
                                 other party perform, or
                                 confirmation that you are expecting
                                 the other party to perform, the
                                 indicated option.
      DON'T (option code) 254    Indicates the demand that the
                                 other party stop performing,
                                 or confirmation that you are no
                                 longer expecting the other party
                                 to perform, the indicated option.
      IAC                 255    Data Byte 255.
!

!TelnetConstantsRFC854 class methodsFor: 'initialization' stamp: 'DanilOsipchuk 7/5/2013 13:05'!
initialize
	"self initialize"
	SGA :=3. "suppress go ahead"
	SE := 240.
	NOP := 241.
	DM := 242.
	BRK := 243.
	IP := 244.
	AO := 245.
	AYT := 246.
	EC := 247.
	EL := 248.
	GA := 249.
	SB := 250.
	WILL := 251.
	WONT := 252.
	DO := 253.
	DONT := 254.
	IAC := 255! !

Object subclass: #TelnetSessionBasic
	instanceVariableNames: 'socket rawMode defaultMatch lastMatchedString'
	classVariableNames: ''
	poolDictionaries: 'TelnetCmdMatchLib TelnetConstantsRFC854'
	category: 'TelnetSessionBasic'!

!TelnetSessionBasic commentStamp: '<historical>' prior: 0!
TelnetSessionBasic is intended for simple automation tasks using telnet protocol.

It provides some convenience remotely comparable to  'expect' scripting command/libraries (#readToMatch:, #cmd:endMatch:).

The implementation uses Socket and hides telnet control flow details (rfc858) from user. Basically it agrees to do SUPPRESS-GO-AHEAD and rejects every other option suggested by remote side, which is barely enough to talk to popular network equipment (see #processReadData:). 






	!

!TelnetSessionBasic class methodsFor: 'examples' stamp: 'DanilOsipchuk 6/18/2013 15:56'!
exampleCiscoLogin
	"self exampleCiscoLogin"
	| ts |
	ts := TelnetSessionBasic new
	telnetLoginCisco: 'user:password!!enablePassword@host'."perform login over opened telnet connection using preconfigured scenario"
	Transcript open.
	Transcript nextPutAll: (ts cmd: 'show users'); flush. "get output of command, stop at 	the cmd prompt"
	ts close.! !

!TelnetSessionBasic class methodsFor: 'examples' stamp: 'DanilOsipchuk 6/18/2013 16:32'!
exampleJuniperLogin
	"self exampleJuniperLogin"
	| ts |
	ts := TelnetSessionBasic new
		telnetLoginJunos: 'danil:danilPwd@o1'. "perform login using preconfigured scenario"
	
	Transcript  open.
	Transcript nextPutAll: (ts cmd: 'show route extensive'); flush. "get output of command"
	ts close.! !

!TelnetSessionBasic class methodsFor: 'examples' stamp: 'DanilOsipchuk 6/18/2013 17:01'!
exampleJunoscriptLogin
	"self exampleJunoscriptLogin"
	| ts |
	ts := TelnetSessionBasic new
		telnetLoginJunoscript:  'danil:danilPwd@o1'. "perform login using preconfigured scenario"
	
	Transcript  open.
	Transcript nextPutAll: (ts rpc: '<get-interface-information/>' ); flush. "get output of command"
	ts close.! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 7/4/2013 19:39'!
close

	socket
		ifNotNil: 
			[socket close.
			socket := nil]! !

!TelnetSessionBasic methodsFor: 'scripting' stamp: 'DanilOsipchuk 6/18/2013 14:45'!
cmd: aString 

	"send aString as a command, use defaultMatch as end of output"
	^self cmd: aString endMatch: defaultMatch! !

!TelnetSessionBasic methodsFor: 'scripting' stamp: 'DanilOsipchuk 6/18/2013 11:17'!
cmd: aString endMatch: aPattern 
	
	"send aString as a command and read output ending by line matching aPattern"
	self writeln: aString.
	^(self readToMatch: aPattern) allButFirst: aString size + 2  "skip the echoed command in the beginning + crlf "! !

!TelnetSessionBasic methodsFor: 'accessing' stamp: 'DanilOsipchuk 6/18/2013 15:52'!
defaultMatch: anObject

	defaultMatch := anObject! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 6/18/2013 16:11'!
doMatch: aStringOrRegexp against: aString.

	"hack some polymorphism in, to perform optimized matches against plain strings"
	(aStringOrRegexp isKindOf: String) ifTrue: [ ^aString = aStringOrRegexp "matcher is a plain string"  ].
	^aStringOrRegexp matches: aString! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 11:41'!
escapeCollection: aCollection

	|  es  |
	es := WriteStream on: (aCollection class new: aCollection size).
	aCollection do: [:byte |
			byte asInteger == IAC ifTrue: [ es nextPut: byte ].
			es nextPut: byte].
	^es contents! !

!TelnetSessionBasic methodsFor: 'initialize-release' stamp: 'DanilOsipchuk 7/5/2013 11:43'!
initialize 

	self rawModeOff! !

!TelnetSessionBasic methodsFor: 'junoscript' stamp: 'DanilOsipchuk 6/18/2013 16:20'!
junoscriptPI

	^'<?xml version="1.0" encoding="us-ascii"?><junoscript version="1.0">'.

! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 10:54'!
nextFrom: aStream

	| byte |
	byte  := aStream next.
	(byte asInteger == IAC)
		ifTrue: 
			[| cmd option |
			cmd := aStream next.
			cmd asInteger == IAC 
				ifTrue: [^cmd]
				ifFalse: 
					[option := aStream next.
					self processCmd: cmd option: option].
			^nil ].

	(byte asInteger == 0) ifTrue:  [^nil]. "as we are in a text mode - should we skip zeroes which confuse upper protocol? on not? let us skip"
	^byte   
! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 6/18/2013 15:53'!
parseLoginString: aString 
	"set login credentials aString:  <loginString>
	 login:password@host:port -- most usual format 
	 login:password!!enablepassword@host:port -- with the enable password for cisco
	 password!!enablePassword@host:port -- cisco, enable password but no login 
	 host:port -- connect "

	| tokens authPart  socketTokens loginInfo    |
	"defaults"
	loginInfo := Dictionary new
		at: #login put: '';
		at: #password put: '';
		at: #secondPassword  put: '';
		at: #host put: '127.0.0.1';
		at: #port put: 23;
		yourself.
		
	tokens := aString subStrings: '@'. socketTokens := #().
	tokens size = 1 ifTrue:  [socketTokens := tokens first subStrings: ':'].
	tokens size = 2 ifTrue:  [ authPart := tokens first.  socketTokens := tokens second subStrings: ':' ].
		
	authPart ifNotNil: 
			[| authTokens pwdTokens |
				authTokens  := authPart subStrings: ':'.
				authTokens size > 1
					ifTrue: [ loginInfo at: #login put: authTokens first.  pwdTokens := authTokens second subStrings: '!!' ] 
					ifFalse: [ pwdTokens := authTokens first subStrings: '!!'  ].
				pwdTokens size > 1
					ifTrue: [ loginInfo at: # password put: pwdTokens first.
							 loginInfo at: # secondPassword put:  pwdTokens second ]
					ifFalse: [ loginInfo at: # password put: pwdTokens first ]].
				
	socketTokens size = 1 ifTrue: [  loginInfo at: #host put: socketTokens first. ].
	socketTokens size = 2	ifTrue: [ loginInfo at: #host put: socketTokens first. 
				 				loginInfo at:  #port put: socketTokens second asNumber ].
		
	^loginInfo! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 11:21'!
processCmd: cmd option: option 
	| reply |
	" agree doing SGA, reject everything else"
	reply := nil.
	option asInteger = SGA 
		ifTrue: 
			["suppress_go_ahead suggested - agree"
			cmd asInteger = DO ifTrue: [reply := ByteArray with: IAC with: WILL with: SGA ]. 
			cmd asInteger = WILL  ifTrue: [reply := ByteArray with: IAC with: DO with: SGA ]]
		ifFalse: 
			[cmd asInteger = DO "refuse"
				ifTrue: [reply :=  ByteArray with: IAC with: WONT with: option asInteger].
			cmd asInteger = WILL "reject"
				ifTrue:  [reply := ByteArray with: IAC with: DONT with: option asInteger].
			((cmd asInteger = WONT) or: [cmd asInteger = DONT]) ifTrue: [ ^self "silently agree" ]].
	reply ifNotNil: [socket  sendData: reply]. 
	
	! !

!TelnetSessionBasic methodsFor: 'private' stamp: 'DanilOsipchuk 7/5/2013 11:56'!
processReadData: aDataFromSocket 


	| rs ws |
	rs := ReadStream on: aDataFromSocket.
	ws := WriteStream on: (aDataFromSocket class new: aDataFromSocket size).
	[rs atEnd] whileFalse: 
			[| char |
			char := self nextFrom: rs.
			char ifNotNil: [ws nextPut: char] ].
	^ws contents! !

!TelnetSessionBasic methodsFor: 'accessing' stamp: 'DanilOsipchuk 7/5/2013 11:22'!
rawModeOff

	rawMode := false! !

!TelnetSessionBasic methodsFor: 'accessing' stamp: 'DanilOsipchuk 7/5/2013 11:22'!
rawModeOn

	rawMode := true! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 7/5/2013 11:26'!
read
	"read ready output from socket"

	| rawData |
	rawData := socket receiveAvailableData.
	^ rawMode
		ifTrue: [ rawData ]
		ifFalse: [ self processReadData: rawData ]! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 6/18/2013 16:54'!
readToMatch: aPattern

	defaultMatch := aPattern.
	^self readToMatchAnywhere: aPattern  waitFor: 6! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 6/18/2013 15:44'!
readToMatchAnywhere: aRegexp waitFor: seconds
	"read data from socket until any line in buffer matches the pattern. 
	 this should be more reliable but a bit slower than #readToMatchLast:waitFor:"

	| currentLines currentContents currentMatchPosition ws timer matchFound |
	timer := seconds * 1000 + Time millisecondClockValue.
	currentLines := #().
	currentContents := ''.
	ws := WriteStream on: (String new: 64).
	matchFound := false.
	[ matchFound ]
		whileFalse: [ 
			| newData |
			Time millisecondClockValue > timer
				ifTrue: [ self error: 'Timeout' ].
			newData := self read.
			newData isEmpty
				ifFalse: [ "newData arrived, update buffer and look for match"
					| lineIndex |
					ws nextPutAll: newData.
					currentContents := ws contents.
					currentLines := currentContents findTokens: String crlf keep: String crlf.
					currentMatchPosition := 0.	"count from end"
					lineIndex := currentLines size.	"look for a match, scan from the end"
					[ matchFound or: [ lineIndex = 0 ] ]
						whileFalse: [ 
							currentMatchPosition := currentMatchPosition + (currentLines at: lineIndex) size.
							(self doMatch: aRegexp against: (currentLines at: lineIndex))
								ifTrue: [ matchFound := true. lastMatchedString :=  (currentLines at: lineIndex).]
								ifFalse: [ lineIndex := lineIndex - 1 ] ].
					Processor yield ] ].	"we can get here only if match did happen, currentMatchPosition is a sum of lengths of strings beginning from match"
	^ currentContents copyFrom: 1 to: currentContents size - currentMatchPosition! !

!TelnetSessionBasic methodsFor: 'read' stamp: 'DanilOsipchuk 6/18/2013 15:44'!
readToMatchLast: aRegexp waitFor: seconds
	"read data from socket until last line in buffer matches the pattern. "

	| lastLine currentContents ws timer  |
	timer := seconds * 1000 + Time millisecondClockValue.
	currentContents := ''. lastLine := ''.
	ws := WriteStream on: (String new: 64).
	[ self doMatch: aRegexp against:  lastLine ]
		whileFalse: [ 
			| newData |
			Time millisecondClockValue > timer
				ifTrue: [ self error: 'Timeout' ].
			newData := self read.
			newData isEmpty
				ifFalse: [ "newData arrived, update buffer and look for match"
					ws nextPutAll: newData.
					currentContents := ws contents.
					lastLine := currentContents lines last.
					Processor yield ] ].	
	lastMatchedString := lastLine.
	^ currentContents copyFrom: 1 to: currentContents size - lastLine size! !

!TelnetSessionBasic methodsFor: 'scripting' stamp: 'DanilOsipchuk 6/18/2013 17:06'!
rpc: anXmlString

	| ws |
	ws := WriteStream on: (String new: anXmlString  size + 10).
	ws
		nextPutAll: '<rpc>';
		nextPutAll: anXmlString;
		nextPutAll: '</rpc>'.
	self write: ws contents.
	^(self readToMatch: RpcReplyStr) , lastMatchedString  "cmd will return output with stripped closing </rpc-reply> add it back here"
		! !

!TelnetSessionBasic methodsFor: 'scripting' stamp: 'DanilOsipchuk 7/5/2013 12:54'!
script: aString endMatch: aPattern 
	
	"send each line as a command and return output of last command. Output of each command should match aPattern"
	| output |
	output := ''.
	aString linesDo: [ :line | output := self cmd: line endMatch: aPattern ].
	^output! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 6/18/2013 15:54'!
telnetLoginCisco: aString

	"remote system is Cisco, send login and password (with username, without enable)."

	| loginInfo |
	loginInfo := self parseLoginString: aString.
	self close. 
	socket := Socket new connectToHostNamed: (loginInfo at: #host) port: (loginInfo at: #port).

	self readToMatch: CiscoLoginRxp. self writeln: (loginInfo at: #login). 
	self readToMatch: CiscoPasswordRxp. self writeln:  (loginInfo at: #password). 
	self readToMatch: CiscoOpPromptRxp. defaultMatch := lastMatchedString.
	self cmd: 'terminal length 0'.
	self cmd: 'terminal width 512'.
	(loginInfo at: #secondPassword) notEmpty ifTrue: [ 
		self writeln: 'enable'.
		self readToMatch: CiscoPasswordRxp. self writeln: (loginInfo at: #secondPassword).
		self readToMatch: CiscoEnaPromptRxp. defaultMatch := lastMatchedString.]! !

!TelnetSessionBasic methodsFor: 'connection' stamp: 'DanilOsipchuk 6/18/2013 15:49'!
telnetLoginJunos: aString

	"remote system is JUNOS, send login and password"
	| loginInfo |
	loginInfo := self parseLoginString: aString.
	self close. 
	socket := Socket new connectToHostNamed: (loginInfo at: #host) port: (loginInfo at: #port).

	self readToMatch: JunLoginRxp. self writeln: (loginInfo at: #login). 
	self readToMatch: JunPasswordRxp. self writeln: (loginInfo at: #password).
	self readToMatch: JunOpPromptRxp.
	self cmd: ('set cli prompt ' , JunCustomPromptStr) endMatch: JunCustomPromptStr. "set a prompt which can be efficiently and reliably matched against"
	self cmd: 'set cli screen-length 0'. 
	self cmd: 'set cli screen-width 0'. 

	
! !

!TelnetSessionBasic methodsFor: 'junoscript' stamp: 'DanilOsipchuk 6/18/2013 17:01'!
telnetLoginJunoscript: aString

	"open telnet to JUNOS and convert it into junoscript session"
	self telnetLoginJunos: aString.
	self writeln: 'junoscript'. self readToMatch: XmlMetaRxp.
	self write: self junoscriptPI.
	self readToMatch: XmlJunoscriptMetaSessionRxp. defaultMatch := RpcReplyStr.
	
! !

!TelnetSessionBasic methodsFor: 'write' stamp: 'DanilOsipchuk 7/5/2013 11:27'!
write: aString
	"send a string to remote system"

	rawMode
		ifTrue: [ socket sendData: aString ]
		ifFalse: [ socket sendData: (self escapeCollection: aString) ]! !

!TelnetSessionBasic methodsFor: 'write' stamp: 'DanilOsipchuk 7/5/2013 11:28'!
writeln: aString

	self write: (aString , String crlf)! !

TestCase subclass: #TelnetSessionBasicTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'TelnetSessionBasic'!

!TelnetSessionBasicTestCase methodsFor: 'tests' stamp: 'DanilOsipchuk 6/18/2013 14:11'!
testParseLoginString1
	| ts info|
	
	ts := TelnetSessionBasic new.
	info := ts parseLoginString: ''.

	self assert: (info at: #host) = '127.0.0.1'.
	self assert: (info at: #port) = 23.
	self assert: (info at: #login) = ''.
	self assert: (info at: #password) = ''.
	self assert: (info at: #secondPassword) = ''! !

!TelnetSessionBasicTestCase methodsFor: 'tests' stamp: 'DanilOsipchuk 6/18/2013 14:10'!
testParseLoginString2
	| ts info|
	
	ts := TelnetSessionBasic new.
	info := ts parseLoginString: 'hostname'.

	self assert: (info at: #host) = 'hostname'.
	self assert: (info at: #port) = 23.
	self assert: (info at: #login) = ''.
	self assert: (info at: #password) = ''.
	self assert: (info at: #secondPassword) = ''! !

!TelnetSessionBasicTestCase methodsFor: 'tests' stamp: 'DanilOsipchuk 6/18/2013 14:11'!
testParseLoginString3
	| ts info|
	
	ts := TelnetSessionBasic new.
	info := ts parseLoginString: 'hostname:88'.

	self assert: (info at: #host) = 'hostname'.
	self assert: (info at: #port) = 88.
	self assert: (info at: #login) = ''.
	self assert: (info at: #password) = ''.
	self assert: (info at: #secondPassword) = ''! !

!TelnetSessionBasicTestCase methodsFor: 'tests' stamp: 'DanilOsipchuk 6/18/2013 14:12'!
testParseLoginString4
	| ts info|
	
	ts := TelnetSessionBasic new.
	info := ts parseLoginString: 'password@hostname:88'.

	self assert: (info at: #host) = 'hostname'.
	self assert: (info at: #port) = 88.
	self assert: (info at: #login) = ''.
	self assert: (info at: #password) = 'password'.
	self assert: (info at: #secondPassword) = ''! !

!TelnetSessionBasicTestCase methodsFor: 'tests' stamp: 'DanilOsipchuk 6/18/2013 14:12'!
testParseLoginString5
	| ts info|
	
	ts := TelnetSessionBasic new.
	info := ts parseLoginString: 'password!!secondPassword@hostname:88'.

	self assert: (info at: #host) = 'hostname'.
	self assert: (info at: #port) = 88.
	self assert: (info at: #login) = ''.
	self assert: (info at: #password) = 'password'.
	self assert: (info at: #secondPassword) = 'secondPassword'! !

!TelnetSessionBasicTestCase methodsFor: 'tests' stamp: 'DanilOsipchuk 6/18/2013 14:12'!
testParseLoginString6
	| ts info|
	
	ts := TelnetSessionBasic new.
	info := ts parseLoginString: 'login:password!!secondPassword@hostname:88'.

	self assert: (info at: #host) = 'hostname'.
	self assert: (info at: #port) = 88.
	self assert: (info at: #login) = 'login'.
	self assert: (info at: #password) = 'password'.
	self assert: (info at: #secondPassword) = 'secondPassword'! !

!TelnetSessionBasicTestCase methodsFor: 'tests' stamp: 'DanilOsipchuk 6/18/2013 14:13'!
testParseLoginString7
	| ts info|
	
	ts := TelnetSessionBasic new.
	info := ts parseLoginString: 'login:password@hostname:88'.

	self assert: (info at: #host) = 'hostname'.
	self assert: (info at: #port) = 88.
	self assert: (info at: #login) = 'login'.
	self assert: (info at: #password) = 'password'.
	self assert: (info at: #secondPassword) = ''! !
TelnetCmdMatchLib initialize!
TelnetConstantsRFC854 initialize!
